###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import copy
import uuid
from autobahn.util import rtime


class CmdRunResult(object):
    def __init__(self, result, duration=None):
        self.result = result
        self.duration = duration

    def __str__(self):
        return 'CmdRunResult(result={}, duration={})'.format(self.result, self.duration)


class Cmd(object):
    def __init__(self):
        self._started = None

    def _pre(self, session):
        if not session:
            raise Exception('not connected')
        self._started = rtime()

    def _post(self, session, result):
        duration = round(1000. * (rtime() - self._started), 1)
        self._started = None
        return CmdRunResult(result, duration)


class CmdGetDomainStatus(Cmd):
    """
    GLOBAL REALM or MREALM: get system status.
    """
    def __init__(self, realm=None):
        Cmd.__init__(self)
        self.realm = realm

    async def run(self, session):
        self._pre(session)
        if self.realm:
            result = await session.call('crossbarfabriccenter.mrealm.get_status')
        else:
            result = await session.call('crossbarfabriccenter.domain.get_status')
        return self._post(session, result)


class CmdGetDomainVersion(Cmd):
    """
    GLOBAL REALM: get domain controller software version.
    """
    def __init__(self, realm=None):
        Cmd.__init__(self)
        self.realm = realm

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.domain.get_version')
        return self._post(session, result)


class CmdGetDomainLicense(Cmd):
    """
    GLOBAL REALM: get domain software stack license.
    """
    def __init__(self, realm=None):
        Cmd.__init__(self)
        self.realm = realm

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.domain.get_license')
        return self._post(session, result)


class CmdPair(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdPairNode(CmdPair):
    """
    GLOBAL REALM: Pair a node to a management realm.
    """
    def __init__(self, realm, pubkey, node_id, authextra=None):
        CmdPair.__init__(self)
        self.realm = realm
        self.pubkey = pubkey
        self.node_id = node_id
        self.authextra = authextra

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.pair_node', self.pubkey, self.realm, self.node_id,
                                    self.authextra)
        return self._post(session, result)


class CmdUnpair(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdUnpairNode(CmdUnpair):
    """
    GLOBAL REALM: Unpair a node currently paired to a management realm.
    """
    def __init__(self, pubkey):
        CmdUnpair.__init__(self)
        self.pubkey = pubkey

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.unpair_node_by_pubkey', self.pubkey)
        return self._post(session, result)


class CmdAdd(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdAddRolePermission(CmdAdd):
    """
    MREALM: Add a permission to a role.
    """
    def __init__(self, role, uri, config):
        """

        :param role:
        :param uri:
        :param config:
        """
        CmdAdd.__init__(self)
        self.role = role
        self.uri = uri
        self.config = config

    async def run(self, session):
        self._pre(session)

        try:
            role_oid = uuid.UUID(self.role)
            role_oid = str(role_oid)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']

        result = await session.call('crossbarfabriccenter.mrealm.arealm.add_role_permission', role_oid, self.uri,
                                    self.config)

        return self._post(session, result)


class CmdAddPrincipal(CmdAdd):
    """
    MREALM: Add a principal to an application realm.
    """
    def __init__(self, arealm, principal, config=None):
        """

        :param arealm:
        :param principal:
        :param config:
        """
        CmdAdd.__init__(self)
        self.arealm = arealm
        self.config = config or {}
        self.config['authid'] = principal

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
            arealm_oid = str(arealm_oid)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']

        role = self.config.get('role', None)
        assert role is not None and type(role) == str

        try:
            role_oid = uuid.UUID(role)
            role_obj = await session.call('crossbarfabriccenter.mrealm.arealm.get_role', role_oid)
        except:
            role_obj = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', role)
            role_oid = role_obj['oid']

        config = copy.deepcopy(self.config)
        del config['role']
        config['role_oid'] = role_oid

        try:
            result = await session.call('crossbarfabriccenter.mrealm.arealm.add_principal', arealm_oid, config)
        except Exception as e:
            print(e)
            raise e

        return self._post(session, result)


class CmdAddPrincipalCredential(CmdAdd):
    """
    MREALM: Add a credential to a principal.
    """
    def __init__(self, arealm, principal, config=None):
        """

        :param arealm:
        :param principal:
        :param config:
        """
        CmdAdd.__init__(self)
        self.arealm = arealm
        self.principal = principal
        self.config = config or {}

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
            arealm_oid = str(arealm_oid)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']

        try:
            principal_oid = uuid.UUID(self.principal)
            principal_oid = str(principal_oid)
        except:
            principal = await session.call('crossbarfabriccenter.mrealm.arealm.get_principal_by_name', arealm_oid,
                                           self.principal)
            principal_oid = principal['oid']

        result = await session.call('crossbarfabriccenter.mrealm.arealm.add_principal_credential', arealm_oid,
                                    principal_oid, self.config)

        return self._post(session, result)


class CmdAddApplicationRealmRole(CmdAdd):
    """
    MREALM: Add a role to an application realm.
    """
    def __init__(self, arealm, role, config=None):
        """

        :param arealm:
        :param role:
        :param config:
        """
        CmdAdd.__init__(self)
        self.arealm = arealm
        self.role = role
        self.config = config

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
            arealm_oid = str(arealm_oid)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']

        try:
            role_oid = uuid.UUID(self.role)
            role_oid = str(role_oid)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']

        result = await session.call('crossbarfabriccenter.mrealm.arealm.add_arealm_role', arealm_oid, role_oid,
                                    self.config)

        return self._post(session, result)


class CmdAddRouterClusterWorkerGroup(CmdAdd):
    """
    MREALM: Add a workergroup to a routercluster.
    """
    def __init__(self, cluster, workergroup, config=None):
        """

        :param cluster: Router cluster name (or object ID).
        :param workergroup:  Workergroup name.
        :param config: Configuration of node cluster association.
        """
        CmdAdd.__init__(self)
        self.cluster = cluster
        self.workergroup = workergroup
        self.config = config

    async def run(self, session):
        self._pre(session)

        try:
            cluster_oid = uuid.UUID(self.cluster)
            cluster_oid = str(cluster_oid)
        except:
            routercluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                               self.cluster)
            cluster_oid = routercluster['oid']

        self.config['name'] = self.workergroup
        self.config['cluster_oid'] = cluster_oid

        workergroup = await session.call('crossbarfabriccenter.mrealm.routercluster.add_routercluster_workergroup',
                                         cluster_oid, self.config)
        # workergroup_oid = workergroup['oid']

        return self._post(session, workergroup)


class CmdAddRouterClusterNode(CmdAdd):
    """
    MREALM: Add a node to a routercluster.
    """
    def __init__(self, routercluster, node, config=None):
        """

        :param routercluster: Router cluster name or object ID.
        :param node:  Node name or object ID.
        :param config: Configuration of node cluster association.
        """
        CmdAdd.__init__(self)
        self.routercluster = routercluster
        self.node = node
        self.config = config

    async def run(self, session):
        self._pre(session)

        try:
            routercluster_oid = uuid.UUID(self.routercluster)
            routercluster_oid = str(routercluster_oid)
        except:
            routercluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                               self.routercluster)
            routercluster_oid = routercluster['oid']

        if self.node == 'all' or self.node is None:
            nodes = await session.call('crossbarfabriccenter.mrealm.get_nodes')
        else:
            nodes = [x.strip() for x in self.node.split(',')]

        result = []
        for _node in nodes:
            try:
                node_oid = uuid.UUID(_node)
                node_oid = str(node_oid)
            except:
                node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
                node_oid = node['oid']

            node_added = await session.call('crossbarfabriccenter.mrealm.routercluster.add_routercluster_node',
                                            routercluster_oid, node_oid, self.config)
            result.append(node_added)
        return self._post(session, result)


class CmdAddWebClusterNode(CmdAdd):
    """
    MREALM: Add a node to a webcluster.
    """
    def __init__(self, webcluster, node, config=None):
        """

        :param webcluster: Web cluster name or object ID.
        :param node:  Node name or object ID.
        :param config: Configuration of node cluster association.
        """
        CmdAdd.__init__(self)
        self.webcluster = webcluster
        self.node = node
        self.config = config

    async def run(self, session):
        self._pre(session)

        try:
            webcluster_oid = uuid.UUID(self.webcluster)
            webcluster_oid = str(webcluster_oid)
        except:
            webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                            self.webcluster)
            webcluster_oid = webcluster['oid']

        if self.node == 'all' or self.node is None:
            nodes = await session.call('crossbarfabriccenter.mrealm.get_nodes')
        else:
            nodes = [x.strip() for x in self.node.split(',')]

        result = []
        for _node in nodes:
            try:
                node_oid = uuid.UUID(_node)
                node_oid = str(node_oid)
            except:
                node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
                node_oid = node['oid']

            node_added = await session.call('crossbarfabriccenter.mrealm.webcluster.add_webcluster_node',
                                            webcluster_oid, node_oid, self.config)
            result.append(node_added)
        return self._post(session, result)


class CmdAddWebClusterService(CmdAdd):
    """
    MREALM: Add a service to a webcluster.
    """
    def __init__(self, webcluster, path, config=None):
        CmdAdd.__init__(self)
        self.webcluster = webcluster
        self.path = path
        self.config = config

    async def run(self, session):
        self._pre(session)
        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)
        result = await session.call('crossbarfabriccenter.mrealm.webcluster.add_webcluster_service', webcluster['oid'],
                                    self.path, self.config)
        return self._post(session, result)


class CmdCreate(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdCreateManagementRealm(CmdCreate):
    """
    GLOBAL REALM: Create a new management realm.
    """
    def __init__(self, realm):
        CmdCreate.__init__(self)
        self.realm = realm

    async def run(self, session):
        self._pre(session)
        mrealm = {'name': self.realm}
        result = await session.call('crossbarfabriccenter.mrealm.create_mrealm', mrealm)
        return self._post(session, result)


class CmdCreateApplicationRealm(CmdCreate):
    """
    MREALM REALM: Create a new application realm.
    """
    def __init__(self, realm):
        CmdCreate.__init__(self)
        self.realm = realm

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.arealm.create_arealm', self.realm)
        return self._post(session, result)


class CmdCreateRole(CmdCreate):
    """
    MREALM REALM: Create a new role (for use with application realms).
    """
    def __init__(self, role):
        CmdCreate.__init__(self)
        self.role = role

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.arealm.create_role', self.role)
        return self._post(session, result)


class CmdCreateRouterCluster(CmdCreate):
    """
    MREALM: Create a routercluster.
    """
    def __init__(self, routercluster):
        CmdCreate.__init__(self)
        self.routercluster = routercluster

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.routercluster.create_routercluster',
                                    self.routercluster)
        return self._post(session, result)


class CmdCreateWebCluster(CmdCreate):
    """
    MREALM: Create a webcluster.
    """
    def __init__(self, webcluster):
        CmdCreate.__init__(self)
        self.webcluster = webcluster

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.webcluster.create_webcluster', self.webcluster)
        return self._post(session, result)


class CmdCreateWebService(CmdCreate):
    """
    MREALM: Create a webservice (within a webcluster).
    """
    def __init__(self, webcluster_oid, webservice):
        CmdCreate.__init__(self)
        self.webcluster_oid = webcluster_oid
        self.webservice = webservice

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.add_webcluster_service',
                                    self.webcluster_oid, self.webservice)

        return self._post(session, result)


class CmdCreateDockerContainer(CmdCreate):
    """
    MREALM: Create a Docker container on a node.
    """
    def __init__(self, node, image, config):
        CmdCreate.__init__(self)
        self.node = node
        self.image = image
        self.config = config

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.create', node['oid'], self.image, self.config)
        return self._post(session, result)


class CmdRemove(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdRemoveRouterClusterNode(CmdRemove):
    def __init__(self, cluster, node):
        CmdRemove.__init__(self)
        self.cluster = cluster
        self.node = node

    async def run(self, session):
        self._pre(session)

        routercluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                           self.cluster)

        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)

        result = await session.call('crossbarfabriccenter.mrealm.routercluster.remove_routercluster_node',
                                    routercluster['oid'], node['oid'])

        return self._post(session, result)


class CmdRemoveRouterClusterWorkerGroup(CmdRemove):
    def __init__(self, cluster, workergroup):
        CmdRemove.__init__(self)
        self.cluster = cluster
        self.workergroup = workergroup

    async def run(self, session):
        self._pre(session)

        cluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                     self.cluster)

        workergroup = await session.call(
            'crossbarfabriccenter.mrealm.routercluster.get_routercluster_workergroup_by_name', self.cluster,
            self.workergroup)

        result = await session.call('crossbarfabriccenter.mrealm.routercluster.remove_routercluster_workergroup',
                                    cluster['oid'], workergroup['oid'])

        return self._post(session, result)


class CmdRemoveWebClusterService(CmdRemove):
    def __init__(self, webcluster, path):
        CmdRemove.__init__(self)
        self.webcluster = webcluster
        self.path = path

    async def run(self, session):
        self._pre(session)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)

        webservice = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_service_by_path',
                                        webcluster['oid'], self.path)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.remove_webcluster_service',
                                    webcluster['oid'], webservice['oid'])

        return self._post(session, result)


class CmdRemoveWebClusterNode(CmdRemove):
    def __init__(self, cluster, node):
        CmdRemove.__init__(self)
        self.cluster = cluster
        self.node = node

    async def run(self, session):
        self._pre(session)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name', self.cluster)

        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.remove_webcluster_node', webcluster['oid'],
                                    node['oid'])

        return self._post(session, result)


class CmdRemoveArealmPrincipal(CmdRemove):
    def __init__(self, arealm, principal):
        CmdRemove.__init__(self)
        self.arealm = arealm
        self.principal = principal

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        try:
            principal_oid = uuid.UUID(self.principal)
        except:
            principal = await session.call('crossbarfabriccenter.mrealm.arealm.get_principal_by_name', self.arealm)
            principal_oid = principal['oid']
        else:
            principal_oid = str(principal_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.remove_principal', arealm_oid, principal_oid)

        return self._post(session, result)


class CmdRemoveArealmPrincipalCredential(CmdRemove):
    def __init__(self, arealm, principal, credential):
        CmdRemove.__init__(self)
        self.arealm = arealm
        self.principal = principal
        self.credential = credential

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        try:
            principal_oid = uuid.UUID(self.principal)
        except:
            principal = await session.call('crossbarfabriccenter.mrealm.arealm.get_principal_by_name', self.arealm)
            principal_oid = principal['oid']
        else:
            principal_oid = str(principal_oid)

        credential_oid = uuid.UUID(self.credential)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.remove_principal_credential', arealm_oid,
                                    principal_oid, credential_oid)

        return self._post(session, result)


class CmdRemoveRolePermission(CmdRemove):
    def __init__(self, role, path):
        CmdRemove.__init__(self)
        self.role = role
        self.path = path

    async def run(self, session):
        self._pre(session)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        permission_oids = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_permissions_by_uri',
                                             role_oid, self.path)

        results = []
        for permission_oid in permission_oids:
            result = await session.call('crossbarfabriccenter.mrealm.arealm.remove_role_permission', role_oid,
                                        permission_oid)
            results.append(result)

        return self._post(session, result)


class CmdRemoveArealmRole(CmdRemove):
    def __init__(self, arealm, role):
        CmdRemove.__init__(self)
        self.arealm = arealm
        self.role = role

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.remove_arealm_role', arealm_oid, role_oid)

        return self._post(session, result)


class CmdDelete(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdDeleteManagementRealm(CmdDelete):
    def __init__(self, realm, cascade):
        CmdDelete.__init__(self)
        self.realm = realm
        self.cascade = cascade

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.delete_mrealm_by_name',
                                    self.realm,
                                    cascade=self.cascade)
        return self._post(session, result)


class CmdDeleteApplicationRealm(CmdDelete):
    def __init__(self, realm, cascade):
        CmdDelete.__init__(self)
        self.realm = realm
        self.cascade = cascade

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.realm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.realm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.delete_arealm', arealm_oid)
        return self._post(session, result)


class CmdDeleteRole(CmdDelete):
    def __init__(self, role):
        CmdDelete.__init__(self)
        self.role = role

    async def run(self, session):
        self._pre(session)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.delete_role', role_oid)
        return self._post(session, result)


class CmdDeleteRouterCluster(CmdDelete):
    """
    MREALM: delete routercluster by UUID.
    """
    def __init__(self, routercluster):
        CmdDelete.__init__(self)
        self.routercluster = routercluster

    async def run(self, session):
        self._pre(session)

        routercluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                           self.routercluster)

        result = await session.call('crossbarfabriccenter.mrealm.routercluster.delete_routercluster',
                                    routercluster['oid'])

        return self._post(session, result)


class CmdDeleteWebCluster(CmdDelete):
    """
    MREALM: delete webcluster by UUID.
    """
    def __init__(self, webcluster):
        CmdDelete.__init__(self)
        self.webcluster = webcluster

    async def run(self, session):
        self._pre(session)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.delete_webcluster', webcluster['oid'])

        return self._post(session, result)


class CmdList(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdListManagementRealms(CmdList):
    """
    GLOBAL REALM: Get list of management realms.
    """
    def __init__(self, names=None):
        CmdList.__init__(self)
        self.names = names

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.list_mrealms', return_names=self.names)
        return self._post(session, result)


class CmdListNodes(CmdList):
    """
    GLOBAL REALM: Get list of nodes in management realms.
    """
    def __init__(self, online=None, offline=None, names=None):
        CmdList.__init__(self)
        self.online = online
        self.offline = offline
        self.names = names

    async def run(self, session):
        self._pre(session)
        status = None
        if self.online is not None or self.offline is not None:
            if self.online:
                status = 'online'
            elif self.offline:
                status = 'offline'
        result = await session.call('crossbarfabriccenter.mrealm.get_nodes', status, return_names=self.names)
        return self._post(session, result)


class CmdListWorkers(CmdList):
    """
    MREALM: Get list of workers on a node.
    """
    def __init__(self, node):
        CmdList.__init__(self)
        self.node = node

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.node.get_workers', node['oid'])
        return self._post(session, result)


class CmdListRouterRealms(CmdList):
    def __init__(self, node, worker):
        CmdList.__init__(self)
        self.node = node
        self.worker = worker

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.remote.router.get_router_realms', self.node, self.worker)
        return self._post(session, result)


class CmdListRouterTransports(CmdList):
    def __init__(self, node, worker):
        CmdList.__init__(self)
        self.node = node
        self.worker = worker

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.remote.router.get_router_transports', self.node, self.worker)
        return self._post(session, result)


class CmdListARealms(CmdList):
    """
    MREALM: Get list of application realms defined on a mrealm.
    """
    def __init__(self, names=False):
        CmdList.__init__(self)
        self.names = names

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.list_arealms', return_names=self.names)
        return self._post(session, result)


class CmdListARealmRoles(CmdList):
    """
    MREALM: Get list of roles associated with the given application realm defined on a mrealm.
    """
    def __init__(self, arealm, names=False):
        CmdList.__init__(self)
        self.arealm = arealm
        self.names = names

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.list_arealm_roles',
                                    arealm_oid,
                                    return_names=self.names)
        return self._post(session, result)


class CmdListRoles(CmdList):
    """
    MREALM: Get list of roles defined on a mrealm.
    """
    def __init__(self, names=False):
        CmdList.__init__(self)
        self.names = names

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.list_roles', return_names=self.names)
        return self._post(session, result)


class CmdListRolePermissions(CmdList):
    """
    MREALM: Get list of permissions defined for a role.
    """
    def __init__(self, role):
        CmdList.__init__(self)
        self.role = role

    async def run(self, session):
        self._pre(session)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.list_role_permissions', role_oid)
        return self._post(session, result)


class CmdListPrincipals(CmdList):
    """
    MREALM: Get list of principals defined on a mrealm.
    """
    def __init__(self, arealm, names=False):
        CmdList.__init__(self)
        self.arealm = arealm
        self.names = names

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.list_principals',
                                    arealm_oid,
                                    return_names=self.names)
        return self._post(session, result)


class CmdListPrincipalCredentials(CmdList):
    """
    MREALM: Get list of credentials of a principal.
    """
    def __init__(self, arealm, principal):
        CmdList.__init__(self)
        self.arealm = arealm
        self.principal = principal

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        try:
            principal_oid = uuid.UUID(self.principal)
        except:
            principal = await session.call('crossbarfabriccenter.mrealm.arealm.get_principal_by_name', arealm_oid,
                                           self.principal)
            principal_oid = principal['oid']
        else:
            principal_oid = str(principal_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.list_principal_credentials', arealm_oid,
                                    principal_oid)

        return self._post(session, result)


class CmdListRouterClusters(CmdList):
    """
    MREALM: Get list of webclusters defined on a mrealm.
    """
    def __init__(self, names=False):
        CmdList.__init__(self)
        self.names = names

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.routercluster.list_routerclusters',
                                    return_names=self.names)
        return self._post(session, result)


class CmdListRouterClusterNodes(CmdList):
    """
    MREALM: Get list of nodes associated with a routercluster.
    """
    def __init__(self, cluster, names=None, filter_status=None):
        CmdList.__init__(self)
        self.cluster = cluster
        self.names = names
        self.filter_status = filter_status

    async def run(self, session):
        self._pre(session)
        cluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                     self.cluster)
        result = await session.call('crossbarfabriccenter.mrealm.routercluster.list_routercluster_nodes',
                                    cluster['oid'],
                                    return_names=self.names,
                                    filter_by_status=self.filter_status)
        return self._post(session, result)


class CmdListRouterClusterWorkerGroups(CmdList):
    """
    MREALM: Get list of workergroups running in a routercluster.
    """
    def __init__(self, cluster, names=None, filter_status=None):
        CmdList.__init__(self)
        self.cluster = cluster
        self.names = names
        self.filter_status = filter_status

    async def run(self, session):
        self._pre(session)

        try:
            cluster_oid = str(uuid.UUID(self.cluster))
        except:
            cluster_obj = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                             self.cluster)
            cluster_oid = cluster_obj['oid']

        workergroups = await session.call('crossbarfabriccenter.mrealm.routercluster.list_routercluster_workergroups',
                                          cluster_oid,
                                          return_names=self.names,
                                          filter_by_status=self.filter_status)

        return self._post(session, workergroups)


class CmdListWebClusters(CmdList):
    """
    MREALM: Get list of webclusters defined on a mrealm.
    """
    def __init__(self, names=False):
        CmdList.__init__(self)
        self.names = names

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.mrealm.webcluster.list_webclusters', return_names=self.names)
        return self._post(session, result)


class CmdListWebClusterNodes(CmdList):
    """
    MREALM: Get list of nodes associated with a webcluster.
    """
    def __init__(self, cluster, names=None, filter_status=None):
        CmdList.__init__(self)
        self.cluster = cluster
        self.names = names
        self.filter_status = filter_status

    async def run(self, session):
        self._pre(session)
        cluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name', self.cluster)
        result = await session.call('crossbarfabriccenter.mrealm.webcluster.list_webcluster_nodes',
                                    cluster['oid'],
                                    return_names=self.names,
                                    filter_by_status=self.filter_status)
        return self._post(session, result)


class CmdListWebClusterWebService(CmdList):
    """
    MREALM: Get list of webcluster-webservices defined on a webcluster.
    """
    def __init__(self, webcluster):
        CmdList.__init__(self)
        self.webcluster = webcluster

    async def run(self, session):
        self._pre(session)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.list_webcluster_services',
                                    webcluster['oid'])
        return self._post(session, result)


class CmdListDockerImages(CmdList):
    """
    MREALM: Get list of Docker images available on a node.
    """
    def __init__(self, node):
        CmdList.__init__(self)
        self.node = node

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.get_images', node['oid'])
        return self._post(session, result)


class CmdListDockerContainers(CmdList):
    """
    MREALM: Get list of Docker containers on a node.
    """
    def __init__(self, node):
        CmdList.__init__(self)
        self.node = node

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.get_containers', node['oid'])
        return self._post(session, result)


class CmdShow(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdShowFabric(CmdShow):
    def __init__(self):
        CmdShow.__init__(self)

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.show_fabric')
        return self._post(session, result)


class CmdShowManagementRealm(CmdShow):
    def __init__(self, realm):
        CmdShow.__init__(self)
        self.realm = realm

    async def run(self, session):
        self._pre(session)

        list_all = False
        if self.realm == 'all' or self.realm is None:
            mrealms = await session.call('crossbarfabriccenter.mrealm.list_mrealms')
        else:
            mrealms = [x.strip() for x in self.realm.split(',')]

        result = []
        for mrealm in mrealms:
            try:
                mrealm_oid = uuid.UUID(mrealm)
            except:
                mrealm_obj = await session.call('crossbarfabriccenter.mrealm.get_mrealm_by_name', mrealm)
            else:
                mrealm_obj = await session.call('crossbarfabriccenter.mrealm.get_mrealm', str(mrealm_oid))
            result.append(mrealm_obj)

        if list_all:
            return self._post(session, result)
        else:
            return self._post(session, result[0])


class CmdShowNode(CmdShow):
    def __init__(self, node=None):
        """
        Get node metadata object.

        :param node: Node ID (a UUID string) or node name or `None`
            to get metadata for all nodes (paired) in the management realm.
        :type node: str

        :returns: List of node metadata objects.
        :rtype: list
        """
        CmdShow.__init__(self)
        self.node = node

    async def run(self, session):
        self._pre(session)

        list_all = False
        if self.node == 'all' or self.node is None:
            nodes = await session.call('crossbarfabriccenter.mrealm.get_nodes')
        else:
            nodes = [x.strip() for x in self.node.split(',')]

        result = []
        for node in nodes:
            try:
                node_oid = uuid.UUID(node)
            except:
                node_obj = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', node)
            else:
                node_obj = await session.call('crossbarfabriccenter.mrealm.get_node', str(node_oid))
            result.append(node_obj)

        if list_all:
            return self._post(session, result)
        else:
            return self._post(session, result[0])


class CmdShowDockerImage(CmdShow):
    def __init__(self, node, image):
        CmdShow.__init__(self)
        self.node = node
        self.image = image

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.get_image', node['oid'], self.image)
        return self._post(session, result)


class CmdShowDockerContainer(CmdShow):
    def __init__(self, node, container):
        CmdShow.__init__(self)
        self.node = node
        self.container = container

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.get_container', node['oid'], self.container)
        return self._post(session, result)


class CmdShowDocker(CmdShow):
    def __init__(self, node, status=True):
        CmdShow.__init__(self)
        self.node = node
        self.status = status

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)

        info = await session.call('crossbarfabriccenter.remote.docker.get_info', node['oid'])
        version = await session.call('crossbarfabriccenter.remote.docker.get_version', node['oid'])

        result = {
            'info': info,
            'version': version,
        }
        if self.status:
            result_ping = await session.call('crossbarfabriccenter.remote.docker.get_ping', node['oid'])
            result_df = await session.call('crossbarfabriccenter.remote.docker.get_df', node['oid'])
            result['status'] = {
                'ping': result_ping,
                'df': result_df,
            }
        return self._post(session, result)


class CmdShowWorker(CmdShow):
    def __init__(self, node, worker):
        CmdShow.__init__(self)
        self.node = node
        self.worker = worker

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.show_worker', self.node, self.worker)
        return self._post(session, result)


class CmdShowTransport(CmdShow):
    def __init__(self, node, worker, transport):
        CmdShow.__init__(self)
        self.node = node
        self.worker = worker
        self.transport = transport

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.show_transport', self.node, self.worker, self.transport)
        return self._post(session, result)


class CmdShowRealm(CmdShow):
    def __init__(self, node, worker, realm):
        CmdShow.__init__(self)
        self.node = node
        self.worker = worker
        self.realm = realm

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.show_realm', self.node, self.worker, self.realm)
        return self._post(session, result)


class CmdShowComponent(CmdShow):
    def __init__(self, node, worker, component):
        CmdShow.__init__(self)
        self.node = node
        self.worker = worker
        self.component = component

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.show_component', self.node, self.worker, self.component)
        return self._post(session, result)


class CmdShowApplicationRealm(CmdShow):
    """
    MREALM: show application realm by UUID or name.
    """
    def __init__(self, arealm):
        CmdShow.__init__(self)
        self.arealm = arealm

    async def run(self, session):
        self._pre(session)

        show_many = False
        if self.arealm == 'all' or self.arealm is None:
            arealms = await session.call('crossbarfabriccenter.mrealm.arealm.list_arealms')
            show_many = True
        else:
            if ',' in self.arealm:
                arealms = [x.strip() for x in self.arealm.split(',')]
            else:
                arealms = [self.arealm]

        result = []
        for arealm in arealms:
            try:
                arealm_oid = uuid.UUID(arealm)
            except:
                arealm_obj = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', arealm)
            else:
                arealm_obj = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm', str(arealm_oid))

            result.append(arealm_obj)

        if show_many:
            return self._post(session, result)
        else:
            return self._post(session, result[0])


class CmdShowPrincipal(CmdShow):
    """
    MREALM: show principal on an application realm role.
    """
    def __init__(self, arealm, principal):
        CmdShow.__init__(self)
        self.arealm = arealm
        self.principal = principal

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        try:
            principal_oid = uuid.UUID(self.principal)
        except:
            principal = await session.call('crossbarfabriccenter.mrealm.arealm.get_principal_by_name', arealm_oid,
                                           self.principal)
        else:
            principal = await session.call('crossbarfabriccenter.mrealm.arealm.get_principal', arealm_oid,
                                           str(principal_oid))

        return self._post(session, principal)


class CmdShowRole(CmdShow):
    """
    MREALM: show role by UUID or name.
    """
    def __init__(self, role):
        CmdShow.__init__(self)
        self.role = role

    async def run(self, session):
        self._pre(session)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.get_role', role_oid)

        return self._post(session, result)


class CmdShowRolePermission(CmdShow):
    """
    MREALM: show role permission by role UUID or name, and optionally URI
    """
    def __init__(self, role, uri):
        CmdShow.__init__(self)
        self.role = role
        self.uri = uri

    async def run(self, session):
        self._pre(session)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_permissions_by_uri', role_oid,
                                    self.uri)

        return self._post(session, result)


class CmdShowARealmRole(CmdShow):
    """
    MREALM: show application realm role association.
    """
    def __init__(self, arealm, role):
        CmdShow.__init__(self)
        self.arealm = arealm
        self.role = role

    async def run(self, session):
        self._pre(session)

        try:
            arealm_oid = uuid.UUID(self.arealm)
        except:
            arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)
            arealm_oid = arealm['oid']
        else:
            arealm_oid = str(arealm_oid)

        try:
            role_oid = uuid.UUID(self.role)
        except:
            role = await session.call('crossbarfabriccenter.mrealm.arealm.get_role_by_name', self.role)
            role_oid = role['oid']
        else:
            role_oid = str(role_oid)

        association = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_role', arealm_oid, role_oid)

        return self._post(session, association)


class CmdShowRouterCluster(CmdShow):
    """
    MREALM: show webcluster by UUID or name.
    """
    def __init__(self, cluster):
        CmdShow.__init__(self)
        self.cluster = cluster

    async def run(self, session):
        self._pre(session)

        show_many = False
        if self.cluster == 'all' or self.cluster is None:
            clusters = await session.call('crossbarfabriccenter.mrealm.routercluster.list_routerclusters')
            show_many = True
        else:
            if ',' in self.cluster:
                clusters = [x.strip() for x in self.cluster.split(',')]
                show_many = True
            else:
                clusters = [self.cluster]

        result = []
        for cluster in clusters:
            try:
                cluster_oid = uuid.UUID(cluster)
            except:
                cluster_obj = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                                 cluster)
            else:
                cluster_oid = str(cluster_oid)
                cluster_obj = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster',
                                                 cluster_oid)

            result.append(cluster_obj)

        if show_many:
            return self._post(session, result)
        else:
            return self._post(session, result[0])


class CmdShowRouterClusterNode(CmdShow):
    """
    MREALM: show routercluster-node by (UUID, UUID).
    """
    def __init__(self, cluster, node):
        CmdShow.__init__(self)
        self.cluster = cluster
        self.node = node

    async def run(self, session):
        self._pre(session)

        try:
            cluster_oid = uuid.UUID(self.cluster)
        except:
            cluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                         self.cluster)
            cluster_oid = cluster['oid']
        else:
            cluster_oid = str(cluster_oid)

        if self.node == 'all' or self.node is None:
            nodes = await session.call('crossbarfabriccenter.mrealm.routercluster.list_routercluster_nodes',
                                       cluster_oid)
        else:
            nodes = [x.strip() for x in self.node.split(',')]

        result = []
        for node in nodes:
            try:
                node_oid = uuid.UUID(node)
            except:
                node_obj = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', node)
                node_oid = node_obj['oid']
            else:
                node_oid = str(node_oid)

            rc_node = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_node',
                                         cluster_oid, node_oid)
            result.append(rc_node)

        return self._post(session, result)


class CmdShowRouterClusterWorkerGroup(CmdShow):
    """
    MREALM: show routercluster workergroup by UUID or name.
    """
    def __init__(self, cluster, workergroup):
        CmdShow.__init__(self)
        self.cluster = cluster
        self.workergroup = workergroup

    async def run(self, session):
        self._pre(session)

        workergroup = await session.call(
            'crossbarfabriccenter.mrealm.routercluster.get_routercluster_workergroup_by_name', self.cluster,
            self.workergroup)

        return self._post(session, workergroup)


class CmdShowWebCluster(CmdShow):
    """
    MREALM: show webcluster by UUID or name.
    """
    def __init__(self, cluster):
        CmdShow.__init__(self)
        self.cluster = cluster

    async def run(self, session):
        self._pre(session)

        list_all = False
        if self.cluster == 'all' or self.cluster is None:
            clusters = await session.call('crossbarfabriccenter.mrealm.webcluster.list_webclusters')
            list_all = True
        else:
            clusters = [x.strip() for x in self.cluster.split(',')]

        result = []
        for cluster in clusters:
            try:
                cluster_oid = uuid.UUID(cluster)
            except:
                cluster_obj = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                                 cluster)
            else:
                cluster_oid = str(cluster_oid)
                cluster_obj = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster', cluster_oid)

            result.append(cluster_obj)

        if list_all:
            return self._post(session, result)
        else:
            return self._post(session, result[0])


class CmdShowWebClusterWebService(CmdShow):
    """
    MREALM: show webcluster-webservice by (UUID, UUID).
    """
    def __init__(self, cluster, path):
        CmdShow.__init__(self)
        self.cluster = cluster
        self.path = path

    async def run(self, session):
        self._pre(session)

        try:
            cluster_oid = uuid.UUID(self.cluster)
        except:
            cluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name', self.cluster)
            cluster_oid = cluster['oid']
        else:
            cluster_oid = str(cluster_oid)

        webservice = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_service_by_path',
                                        cluster_oid, self.path)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_service', cluster_oid,
                                    webservice['oid'])

        return self._post(session, result)


class CmdShowWebClusterNode(CmdShow):
    """
    MREALM: show webcluster-node by (UUID, UUID).
    """
    def __init__(self, cluster, node):
        CmdShow.__init__(self)
        self.cluster = cluster
        self.node = node

    async def run(self, session):
        self._pre(session)

        try:
            cluster_oid = uuid.UUID(self.cluster)
        except:
            cluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name', self.cluster)
            cluster_oid = cluster['oid']
        else:
            cluster_oid = str(cluster_oid)

        if self.node == 'all' or self.node is None:
            nodes = await session.call('crossbarfabriccenter.mrealm.webcluster.list_webcluster_nodes', cluster_oid)
        else:
            nodes = [x.strip() for x in self.node.split(',')]

        result = []
        for node in nodes:
            try:
                node_oid = uuid.UUID(node)
            except:
                node_obj = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', node)
                node_oid = node_obj['oid']
            else:
                node_oid = str(node_oid)

            cluster_node = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_node',
                                              cluster_oid, node_oid)
            result.append(cluster_node)

        return self._post(session, result)


class CmdStart(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdStartDockerContainer(CmdStart):
    def __init__(self, node, container):
        CmdStart.__init__(self)
        self.node = node
        self.container = container

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.start', node['oid'], self.container)
        return self._post(session, result)


class CmdStartWorker(CmdStart):
    def __init__(self, node_id, worker_id, worker_type, worker_options=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.worker_type = worker_type
        self.worker_options = worker_options

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.remote.node.start_worker',
                                    node_id=self.node_id,
                                    worker_id=self.worker_id,
                                    worker_type=self.worker_type,
                                    worker_options=self.worker_options)
        return self._post(session, result)


class CmdStartContainerWorker(CmdStart):
    def __init__(self, node_id, worker_id, process_title=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.process_title = process_title

    async def run(self, session):
        self._pre(session)

        options = {}
        if self.process_title:
            options['title'] = self.process_title

        result = await session.call('crossbarfabriccenter.remote.node.start_worker',
                                    node_id=self.node_id,
                                    worker_id=self.worker_id,
                                    worker_type='container',
                                    worker_options=options)
        return self._post(session, result)


class CmdStartContainerComponent(CmdStart):
    def __init__(self, node_id, worker_id, component_id, config=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.component_id = component_id

        if config:
            self.config = config
        else:
            self.config = {
                'type': 'class',
                'transport': {
                    'type': None,
                    'endpoint': {
                        'type': 'websocket',
                        'url': 'ws://localhost:8080/ws'
                    }
                }
            }

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.remote.container.start_component',
                                    node_id=self.node_id,
                                    worker_id=self.worker_id,
                                    component_id=self.component_id,
                                    config=self.config)
        return self._post(session, result)


class CmdStartRouterWorker(CmdStart):
    def __init__(self, node_id, worker_id, process_title=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.process_title = process_title

    async def run(self, session):
        self._pre(session)

        options = {}
        if self.process_title:
            options['title'] = self.process_title

        result = await session.call('crossbarfabriccenter.remote.node.start_worker',
                                    node_id=self.node_id,
                                    worker_id=self.worker_id,
                                    worker_type='router',
                                    worker_options=options)
        return self._post(session, result)


class CmdStartRouterRealm(CmdStart):
    def __init__(self, node_id, worker_id, realm_id, config=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.realm_id = realm_id

        if config:
            self.config = config
        else:
            self.config = {
                "options": {
                    "enable_meta_api": True,
                    "bridge_meta_api": False
                },
                "roles": [{
                    "name":
                    "anonymous",
                    "permissions": [{
                        "uri": "",
                        "match": "prefix",
                        "allow": {
                            "call": True,
                            "register": True,
                            "publish": True,
                            "subscribe": True
                        },
                        "disclose": {
                            "caller": False,
                            "publisher": False
                        },
                        "cache": True
                    }]
                }]
            }

        if 'name' not in self.config:
            self.config['name'] = self.realm_id

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.remote.router.start_router_realm', self.node_id,
                                    self.worker_id, self.realm_id, self.config)
        return self._post(session, result)


class CmdStartRouterTransport(CmdStart):
    def __init__(self, node_id, worker_id, transport_id, config=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.transport_id = transport_id
        if config:
            self.config = config
        else:
            self.config = {
                "type": "universal",
                "endpoint": {
                    "type": "tcp",
                    "port": 8080
                },
                "rawsocket": {},
                "websocket": {
                    "ws": {
                        "type": "websocket"
                    }
                },
                "web": {
                    "paths": {
                        "/": {
                            "type": "nodeinfo"
                        }
                    }
                }
            }

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.remote.router.start_router_transport', self.node_id,
                                    self.worker_id, self.transport_id, self.config)
        return self._post(session, result)


class CmdStartWebTransportService(CmdStart):
    def __init__(self, node_id, worker_id, transport_id, path, config=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.transport_id = transport_id
        self.path = path
        if config:
            self.config = config
        else:
            self.config = {
                "type": "info",
            }

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.remote.router.start_web_transport_service', self.node_id,
                                    self.worker_id, self.transport_id, self.path, self.config)
        return self._post(session, result)


class CmdStartGuestWorker(CmdStart):
    def __init__(self, node_id, worker_id, config=None):
        CmdStart.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        if config:
            self.config = config
        else:
            self.config = {
                "type": "guest",
                "executable": "/bin/date",
                "arguments": [],
                "options": {
                    "workdir": "..",
                    "env": {
                        "inherit": True
                    }
                }
            }

    async def run(self, session):
        self._pre(session)

        result = await session.call('crossbarfabriccenter.remote.node.start_worker', self.node_id, self.worker_id,
                                    'guest', self.config)
        return self._post(session, result)


class CmdStartRouterCluster(CmdStart):
    def __init__(self, routercluster):
        CmdStart.__init__(self)
        self.routercluster = routercluster

    async def run(self, session):
        self._pre(session)

        routercluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                           self.routercluster)

        result = await session.call('crossbarfabriccenter.mrealm.routercluster.start_routercluster',
                                    routercluster['oid'])
        return self._post(session, result)


class CmdStartApplicationRealm(CmdStart):
    def __init__(self, arealm, routercluster, workergroup, webcluster):
        CmdStart.__init__(self)
        self.arealm = arealm
        self.routercluster = routercluster
        self.workergroup = workergroup
        self.webcluster = webcluster

    async def run(self, session):
        self._pre(session)

        arealm = await session.call('crossbarfabriccenter.mrealm.arealm.get_arealm_by_name', self.arealm)

        workergroup = await session.call(
            'crossbarfabriccenter.mrealm.routercluster.get_routercluster_workergroup_by_name', self.routercluster,
            self.workergroup)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)

        result = await session.call('crossbarfabriccenter.mrealm.arealm.start_arealm', arealm['oid'],
                                    workergroup['oid'], webcluster['oid'])
        return self._post(session, result)


class CmdStartWebCluster(CmdStart):
    def __init__(self, webcluster):
        CmdStart.__init__(self)
        self.webcluster = webcluster

    async def run(self, session):
        self._pre(session)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.start_webcluster', webcluster['oid'])
        return self._post(session, result)


class CmdStop(Cmd):
    def __init__(self):
        Cmd.__init__(self)


class CmdStopDockerContainer(CmdStop):
    def __init__(self, node, container):
        CmdStop.__init__(self)
        self.node = node
        self.container = container

    async def run(self, session):
        self._pre(session)
        node = await session.call('crossbarfabriccenter.mrealm.get_node_by_authid', self.node)
        result = await session.call('crossbarfabriccenter.remote.docker.stop', node['oid'], self.container)
        return self._post(session, result)


class CmdStopWorker(CmdStop):
    def __init__(self, node_id, worker_id):
        CmdStop.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.remote.node.stop_worker',
                                    node_id=self.node_id,
                                    worker_id=self.worker_id)
        return self._post(session, result)


class CmdStopRouterRealm(CmdStop):
    def __init__(self, node_id, worker_id, realm_id):
        CmdStop.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.realm_id = realm_id

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.remote.router.stop_router_realm', self.node_id,
                                    self.worker_id, self.realm_id)
        return self._post(session, result)


class CmdStopRouterTransport(CmdStop):
    def __init__(self, node_id, worker_id, transport_id):
        CmdStop.__init__(self)
        self.node_id = node_id
        self.worker_id = worker_id
        self.transport_id = transport_id

    async def run(self, session):
        self._pre(session)
        result = await session.call('crossbarfabriccenter.remote.router.stop_router_transport', self.node_id,
                                    self.worker_id, self.transport_id)
        return self._post(session, result)


class CmdStopRouterCluster(CmdStop):
    def __init__(self, routercluster):
        CmdStop.__init__(self)
        self.routercluster = routercluster

    async def run(self, session):
        self._pre(session)

        routercluster = await session.call('crossbarfabriccenter.mrealm.routercluster.get_routercluster_by_name',
                                           self.routercluster)

        result = await session.call('crossbarfabriccenter.mrealm.routercluster.stop_routercluster',
                                    routercluster['oid'])

        return self._post(session, result)


class CmdStopWebCluster(CmdStop):
    def __init__(self, webcluster):
        CmdStop.__init__(self)
        self.webcluster = webcluster

    async def run(self, session):
        self._pre(session)

        webcluster = await session.call('crossbarfabriccenter.mrealm.webcluster.get_webcluster_by_name',
                                        self.webcluster)

        result = await session.call('crossbarfabriccenter.mrealm.webcluster.stop_webcluster', webcluster['oid'])

        return self._post(session, result)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from crossbar.edge.node import FabricNode, FabricNodeControllerSession


class XbrNetworkNodeControllerSession(FabricNodeControllerSession):
    pass


class XbrNetworkNode(FabricNode):
    """
    Crossbar.io node personality.
    """
    DEFAULT_CONFIG_PATH = 'network/node/config.json'
    NODE_CONTROLLER = XbrNetworkNodeControllerSession



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
import json
import uuid
import datetime
from pprint import pformat
from typing import Optional, Tuple, Dict, Any, List, Set

from http import cookies as http_cookies

import numpy as np

import zlmdb

import txaio

txaio.use_twisted()  # noqa
from txaio import make_logger, time_ns

from autobahn import util
from autobahn.util import hltype, hlval, hlid

from autobahn.wamp.interfaces import ISession
# from crossbar.router.protocol import WampWebSocketServerProtocol

from cfxdb import cookiestore

__all__ = (
    'CookieStore',
    'CookieStoreMemoryBacked',
    'CookieStoreFileBacked',
    'CookieStoreDatabaseBacked',
)


class CookieStore(object):
    """
    Cookie store common base.
    """

    log = make_logger()

    def __init__(self, config: Dict[str, Any]):
        """

        .. code-block:: json

            {
                "type": "memory",
                "name": "cbtid",
                "length": 24,
                "max_age": 604800
            }

        :param config: The cookie configuration.
        """
        self._config = config

        # name of the HTTP cookie in use
        self._cookie_id_field = config.get('name', 'cbtid')

        # length of the cookie (random) ID value
        self._cookie_id_field_length = int(config.get('length', 24))

        # lifetime of the cookie in seconds (http://tools.ietf.org/html/rfc6265#page-20)
        self._cookie_max_age = int(config.get('max_age', 86400 * 7))

        # transient sessions per cookie database
        self._connections: Dict[str, Set[ISession]] = {}

    def parse(self, headers):
        """
        Parse HTTP request header for cookie. If cookie is found, return cookie ID,
        else return None.
        """
        self.log.debug('{func} parsing cookie from {headers}', func=hltype(self.parse), headers=headers)

        # see if there already is a cookie set in the HTTP request
        if 'cookie' in headers:
            try:
                cookie = http_cookies.SimpleCookie()
                header_cookie = str(headers['cookie'])
                cookie.load(header_cookie)

                if self._cookie_id_field not in cookie and self._cookie_id_field in header_cookie:
                    # Sometimes Python can't parse cookie. So let's parse it manually.
                    header_cookies_as_array = header_cookie.split(";")
                    if len(header_cookies_as_array) != 0:
                        header_cookie_id_indexes = []
                        for cookie_raw in header_cookies_as_array:
                            if (self._cookie_id_field + "=") in cookie_raw:
                                header_cookie_id_indexes.append(header_cookies_as_array.index(cookie_raw))
                        if len(header_cookie_id_indexes) > 0:
                            cookie.load(header_cookies_as_array[header_cookie_id_indexes[0]])
            except http_cookies.CookieError:
                pass
            else:
                if self._cookie_id_field in cookie:
                    cbtid = cookie[self._cookie_id_field].value
                    return cbtid
        return None

    def exists(self, cbtid: str) -> bool:
        return False

    def addProto(self, cbtid: str, proto: ISession) -> int:
        """
        Add given WebSocket connection to the set of connections associated
        with the cookie having the given ID. Return the new count of
        connections associated with the cookie.
        """
        self.log.debug('{func} adding proto {proto} for cookie "{cbtid}"',
                       func=hltype(self.addProto),
                       proto=proto,
                       cbtid=hlid(cbtid))

        if self.exists(cbtid):
            if cbtid not in self._connections:
                self._connections[cbtid] = set()
            self._connections[cbtid].add(proto)
            return len(self._connections[cbtid])
        else:
            if cbtid in self._connections:
                del self._connections[cbtid]
            return 0

    def dropProto(self, cbtid: str, proto: ISession):
        """
        Remove given WebSocket connection from the set of connections associated
        with the cookie having the given ID. Return the new count of
        connections associated with the cookie.
        """
        self.log.debug('{func} removing proto {proto} from cookie "{cbtid}"',
                       func=hltype(self.dropProto),
                       proto=proto,
                       cbtid=hlid(cbtid))

        if self.exists(cbtid):
            if cbtid in self._connections:
                # remove this WebSocket connection from the set of connections
                # associated with the same cookie
                self._connections[cbtid].discard(proto)
                remaining = len(self._connections[cbtid])
                if remaining:
                    return remaining
                else:
                    del self._connections[cbtid]
        else:
            if cbtid in self._connections:
                del self._connections[cbtid]
        return 0

    def getProtos(self, cbtid: str) -> List[ISession]:
        """
        Get all WebSocket connections currently associated with the cookie.
        """
        if self.exists(cbtid):
            if cbtid in self._connections:
                return list(self._connections[cbtid])
        return []


class CookieStoreMemoryBacked(CookieStore):
    """
    Memory-backed cookie store.
    """
    def __init__(self, config: Dict[str, Any]):
        CookieStore.__init__(self, config)

        # transient cookie database
        self._cookies: Dict[str, Dict[str, Any]] = {}

        self.log.info('{func} cookie stored created with config=\n{config}',
                      func=hltype(CookieStoreMemoryBacked.__init__),
                      config=pformat(config))

    def exists(self, cbtid):
        """
        Check if cookie with given ID exists.
        """
        cookie_exists = cbtid in self._cookies
        self.log.debug("Cookie {cbtid} exists = {cookie_exists}", cbtid=cbtid, cookie_exists=cookie_exists)
        return cookie_exists

    def create(self):
        """
        Create a new cookie, returning the cookie ID and cookie header value.
        """
        # http://tools.ietf.org/html/rfc6265#page-20
        # 0: delete cookie
        # -1: preserve cookie until browser is closed

        cbtid = util.newid(self._cookie_id_field_length)

        # cookie tracking data
        cbtData = {
            # cookie ID
            'cbtid': cbtid,

            # UTC timestamp when the cookie was created
            'created': util.utcnow(),

            # maximum lifetime of the tracking/authenticating cookie
            'max_age': self._cookie_max_age,

            # when a cookie has been set, and the WAMP session
            # was successfully authenticated thereafter, the latter
            # auth info is store here
            'authid': None,
            'authrole': None,
            'authrealm': None,
            'authmethod': None,
            'authextra': None,
        }

        self._cookies[cbtid] = cbtData

        # set of WAMP transports (WebSocket connections) this
        # cookie is currently used on
        self._connections[cbtid] = set()

        self.log.debug("New cookie {cbtid} created", cbtid=cbtid)

        # do NOT add the "secure" cookie attribute! "secure" refers to the
        # scheme of the Web page that triggered the WS, not WS itself!!
        #
        return cbtid, '%s=%s;max-age=%d' % (self._cookie_id_field, cbtid, cbtData['max_age'])

    def getAuth(self, cbtid):
        """
        Return `(authid, authrole, authmethod, authrealm, authextra)` tuple given cookie ID.
        """
        if cbtid in self._cookies:
            c = self._cookies[cbtid]
            cookie_auth_info = c['authid'], c['authrole'], c['authmethod'], c['authrealm'], c['authextra']
        else:
            cookie_auth_info = None, None, None, None, None

        self.log.debug("Cookie auth info for {cbtid} retrieved: {cookie_auth_info}",
                       cbtid=cbtid,
                       cookie_auth_info=cookie_auth_info)

        return cookie_auth_info

    def setAuth(self, cbtid, authid, authrole, authmethod, authextra, authrealm):
        """
        Set `(authid, authrole, authmethod, authextra)` for given cookie ID.
        """
        if cbtid in self._cookies:
            c = self._cookies[cbtid]
            c['authid'] = authid
            c['authrole'] = authrole
            c['authrealm'] = authrealm
            c['authmethod'] = authmethod
            c['authextra'] = authextra


class CookieStoreFileBacked(CookieStoreMemoryBacked):
    """
    A persistent, file-backed cookie store.

    This cookie store is backed by a file, which is written to in append-only mode.
    Hence, the file is "growing forever". Whenever information attached to a cookie
    is changed (such as a previously anonymous cookie is authenticated), a new cookie
    record is appended. When the store is booting, the file is sequentially scanned.
    The last record for a given cookie ID is remembered in memory.
    """
    def __init__(self, cookie_file_name, config):
        CookieStoreMemoryBacked.__init__(self, config)

        self._cookie_file_name = cookie_file_name

        if not os.path.isfile(self._cookie_file_name):
            self.log.debug("File-backed cookie store created")
        else:
            self.log.debug("File-backed cookie store already exists")

        self._cookie_file = open(self._cookie_file_name, 'a')

        # initialize cookie database
        self._init_store()

        if config['store'].get('purge_on_startup', False):
            self._clean_cookie_file()

    def _iter_persisted(self):
        with open(self._cookie_file_name, 'r') as f:
            for c in f.readlines():
                d = json.loads(c)

                # we do not persist the connections
                # here make sure the cookie loaded has a
                # default connections key to avoid key errors
                # other keys that aren't persisted should be set here
                d['connections'] = set()

                yield d

    def _persist(self, cbtid, c, status='created'):

        self._cookie_file.write(
            json.dumps({
                'cbtid': cbtid,
                status: c['created'],
                'max_age': c['max_age'],
                'authid': c['authid'],
                'authrole': c['authrole'],
                'authmethod': c['authmethod'],
                'authrealm': c['authrealm'],
                'authextra': c['authextra'],
            }) + '\n')
        self._cookie_file.flush()
        os.fsync(self._cookie_file.fileno())

    def _init_store(self):
        n = 0
        for cookie in self._iter_persisted():
            cbtid = cookie.pop('cbtid')
            if cbtid not in self._cookies:
                self._cookies[cbtid] = {}
            self._cookies[cbtid].update(cookie)
            n += 1

        self.log.info("Loaded {cnt_cookie_records} cookie records from file. Cookie store has {cnt_cookies} entries.",
                      cnt_cookie_records=n,
                      cnt_cookies=len(self._cookies))

    def create(self):
        cbtid, header = CookieStoreMemoryBacked.create(self)

        c = self._cookies[cbtid]

        self._persist(cbtid, c)

        self.log.debug("Cookie {cbtid} stored", cbtid=cbtid)

        return cbtid, header

    def setAuth(self, cbtid, authid, authrole, authmethod, authextra, authrealm):

        if self.exists(cbtid):

            cookie = self._cookies[cbtid]

            # only set the changes and write them to the file if any of the values changed
            if (authid != cookie['authid'] or authrole != cookie['authrole'] or authmethod != cookie['authmethod']
                    or authrealm != cookie['authrealm'] or authextra != cookie['authextra']):
                CookieStoreMemoryBacked.setAuth(self, cbtid, authid, authrole, authmethod, authextra, authrealm)
                self._persist(cbtid, cookie, status='modified')

    def _clean_cookie_file(self):
        with open(self._cookie_file_name, 'w') as cookie_file:
            for cbtid, cookie in self._cookies.items():
                expiration_delta = datetime.timedelta(seconds=int(cookie['max_age']))
                upper_limit = util.utcstr(datetime.datetime.now() - expiration_delta)
                if cookie['created'] < upper_limit:
                    # This cookie is expired, discard
                    continue

                cookie_record = json.dumps({
                    'cbtid': cbtid,
                    'created': cookie['created'],
                    'max_age': cookie['max_age'],
                    'authid': cookie['authid'],
                    'authrole': cookie['authrole'],
                    'authmethod': cookie['authmethod'],
                    'authrealm': cookie['authrealm'],
                    'authextra': cookie['authextra'],
                }) + '\n'
                cookie_file.write(cookie_record)

            cookie_file.flush()
            os.fsync(cookie_file.fileno())


class CookieStoreDatabaseBacked(CookieStore):
    """
    A persistent, database-backed cookie store. This implementation uses a zLMDB
    based embedded database with Flatbuffers data serialization.
    """
    def __init__(self, dbpath: str, config: Dict[str, Any]):
        """
        Initialize a database-backed cookiestore. Example configuration:

        .. code-block:: json

            {
                "type": "database",
                "path": ".cookies",
                "purge_on_startup": false,
                "maxsize": 1048576,
                "readonly": false,
                "sync": true
            }

        :param dbpath: Filesystem path to database.
        :param config: Database cookie store configuration.
        """
        self.log.info('{func}: initializing database-backed cookiestore with config=\n{config}',
                      func=hltype(CookieStoreDatabaseBacked.__init__),
                      config=pformat(config))
        CookieStore.__init__(self, config)

        maxsize = config['store'].get('maxsize', 1024 * 2**20)
        assert type(maxsize) == int, "maxsize must be an int, was {}".format(type(maxsize))
        # allow maxsize 128kiB to 128GiB
        assert maxsize >= 128 * 1024 and maxsize <= 128 * 2**30, "maxsize must be >=128kiB and <=128GiB, was {}".format(
            maxsize)

        readonly = config['store'].get('readonly', False)
        assert type(readonly) == bool, "readonly must be a bool, was {}".format(type(readonly))

        sync = config['store'].get('sync', True)
        assert type(sync) == bool, "sync must be a bool, was {}".format(type(sync))

        if config['store'].get('purge_on_startup', False):
            zlmdb.Database.scratch(dbpath)
            self.log.warn('{func}: scratched embedded database (purge_on_startup is enabled)!',
                          func=hltype(CookieStoreDatabaseBacked.__init__))

        self._db = zlmdb.Database(dbpath=dbpath, maxsize=maxsize, readonly=readonly, sync=sync)
        # self._db.__enter__()
        self._schema = cookiestore.CookieStoreSchema.attach(self._db)

        dbstats = self._db.stats(include_slots=True)

        self.log.info('{func}: database-backed cookiestore opened from dbpath="{dbpath}" - dbstats=\n{dbstats}',
                      func=hltype(CookieStoreDatabaseBacked.__init__),
                      dbpath=hlval(dbpath),
                      dbstats=pformat(dbstats))

    def exists(self, cbtid: str) -> bool:
        """
        Check if a cookie with given value currently exists in the cookie store.

        :param cbtid: Cookie value (ID) to check.
        :return: Flag indicating whether a cookie (authenticated or not) is stored in the database.
        """
        # check if a cookie with the given value exists
        with self._db.begin() as txn:
            cookie_exists = self._schema.idx_cookies_by_value[txn, cbtid] is not None
        self.log.debug('{func}(cbtid="{cbtid}") -> {cookie_exists}',
                       func=hltype(self.exists),
                       cbtid=hlval(cbtid),
                       cookie_exists=hlval(cookie_exists))
        return cookie_exists

    def create(self) -> Tuple[str, str]:
        """
        Create and store a new cookie (unauthenticated) in the database.

        :return: A pair with Cookie value (ID, ``cbtid``) and complete Cookie HTTP header value.
        """
        # create new cookie
        cookie = cookiestore.Cookie()
        cookie.oid = uuid.uuid4()
        cookie.created = np.datetime64(time_ns(), 'ns')
        cookie.max_age = self._cookie_max_age
        cookie.name = self._cookie_id_field
        cookie.value = util.newid(self._cookie_id_field_length)

        # write cookie to database
        with self._db.begin(write=True) as txn:
            self._schema.cookies[txn, cookie.oid] = cookie

        self.log.info('{func} new cookie {cbtid} stored in database', func=hltype(self.create), cbtid=cookie.value)

        return cookie.value, '%s=%s;max-age=%d' % (cookie.name, cookie.value, cookie.max_age)

    def getAuth(
            self,
            cbtid: str) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str], Optional[Dict[str, Any]]]:
        """
        Get the authentication info (if any) for a cookie stored in the database.

        :param cbtid: Cookie value (ID) to get authentication for.
        :return: A tuple ``(authid, authrole, authmethod, authrealm, authextra)`` if the cookie exists,
            or a tuple ``(None, None, None, None, None)`` if no cookie with given ``cbtid`` is
            currently stored.
        """
        with self._db.begin() as txn:
            # find cookie OID by cookie value
            cookie_oid = self._schema.idx_cookies_by_value[txn, cbtid]
            if cookie_oid:
                # if we found a cookie OID, read the actual cookie from database
                cookie = self._schema.cookies[txn, cookie_oid]
                assert cookie
                cbtid_ = cookie.value
                cookie_auth_info = cookie.authid, cookie.authrole, cookie.authmethod, cookie.authrealm, cookie.authextra
            else:
                cbtid_ = None
                cookie_auth_info = None, None, None, None, None

        if cbtid_:
            self.log.info('{func} cookie auth info for "{cbtid}" retrieved: {cookie_auth_info}',
                          func=hltype(self.getAuth),
                          cbtid=hlid(cbtid_),
                          cookie_auth_info=cookie_auth_info)
        else:
            self.log.info('{func} no cookie for "{cbtid}" stored in cookiestore database',
                          func=hltype(self.getAuth),
                          cbtid=hlid(cbtid))

        return cookie_auth_info

    def setAuth(self, cbtid: str, authid: Optional[str], authrole: Optional[str], authmethod: Optional[str],
                authextra: Optional[Dict[str, Any]], authrealm: Optional[str]) -> bool:
        """
        Update the authentication information associated and stored for an existing cookie (if any).

        :param cbtid: Cookie value (ID) to update authentication for.
        :param authid: The WAMP authid a cookie-authenticating session is to be assigned.
        :param authrole: The WAMP authrole a cookie-authenticating session is to join under.
        :param authmethod: The WAMP authentication method to be returned to the client performing
            this cookie-based authentication.
        :param authextra: The WAMP authentication extra data to be returned to the client performing
            this cookie-based authentication.
        :param authrealm: The WAMP realm a cookie-authenticating session is to join.
        :return: Flag indicating an existing cookie was modified.
        """

        was_existing = False
        was_modified = False
        with self._db.begin(write=True) as txn:
            cookie_oid = self._schema.idx_cookies_by_value[txn, cbtid]
            if cookie_oid:
                # read current cookie from database
                cookie = self._schema.cookies[txn, cookie_oid]
                assert cookie
                was_existing = True
                if (authid != cookie.authid or authrole != cookie.authrole or authmethod != cookie.authmethod
                        or authrealm != cookie.authrealm or authextra != cookie.authextra):
                    cookie.authid = authid
                    cookie.authrole = authrole
                    cookie.authmethod = authmethod
                    cookie.authrealm = authrealm
                    cookie.authextra = authextra
                    # write updated cookie to database
                    self._schema.cookies[txn, cookie.oid] = cookie
                    was_modified = True

        if was_existing:
            if was_modified:
                self.log.info(
                    '{func} cookie with cbtid="{cbtid}" exists, but was updated (authid="{authid}", authrole='
                    '"{authrole}", authmethod="{authmethod}", authrealm="{authrealm}", authextra={authextra})',
                    func=hltype(self.setAuth),
                    cbtid=hlid(cbtid),
                    authid=hlid(authid),
                    authrole=hlid(authrole),
                    authmethod=hlval(authmethod),
                    authrealm=hlval(authrealm),
                    authextra=pformat(authextra))
            else:
                self.log.info('{func} cookie with cbtid="{cbtid}" exists and needs no update',
                              func=hltype(self.setAuth),
                              cbtid=hlid(cbtid))
        else:
            self.log.info('{func} no cookie to modify with cbtid="{cbtid}" exists',
                          func=hltype(self.setAuth),
                          cbtid=hlid(cbtid))

        return was_modified



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from crossbar.edge.worker.monitor._self import SelfMonitor
from crossbar.edge.worker.monitor._hardware import HWMonitor
from crossbar.edge.worker.monitor._network import NetMonitor
from crossbar.edge.worker.monitor._storage import StorageMonitor
from crossbar.edge.worker.monitor._system import SystemMonitor
from crossbar.edge.worker.monitor._memory import MemoryMonitor
from crossbar.edge.worker.monitor._cpu import CPUMonitor
from crossbar.edge.worker.monitor._disk import IOMonitor
from crossbar.edge.worker.monitor._process import ProcessMonitor

__all__ = ('MONITORS', )

# class map: monitor name -> monitor class
MONITORS = {
    SelfMonitor.ID: SelfMonitor,
    HWMonitor.ID: HWMonitor,
    NetMonitor.ID: NetMonitor,
    StorageMonitor.ID: StorageMonitor,
    SystemMonitor.ID: SystemMonitor,
    MemoryMonitor.ID: MemoryMonitor,
    CPUMonitor.ID: CPUMonitor,
    IOMonitor.ID: IOMonitor,
    ProcessMonitor.ID: ProcessMonitor,
}



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
from txaio import make_logger

try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler
except ImportError:
    HAS_FS_WATCHER = False
else:
    HAS_FS_WATCHER = True

__all__ = ('FilesystemWatcher', 'HAS_FS_WATCHER')

if HAS_FS_WATCHER:

    class FilesystemWatcher:
        """
        Watches a directories for file system changes.
        """

        log = make_logger()

        def __init__(self, working_dir='.', watched_dirs=['.']):
            """

            :param watched_dirs: Directories to watch for changes.
            :type watched_dirs: list of str
            """
            self._working_dir = working_dir
            self._watched_dirs = watched_dirs
            self._started = False
            self._observer = Observer()
            self._handler = FileSystemEventHandler()
            for path in watched_dirs:
                path = os.path.abspath(os.path.join(working_dir, path))
                self._observer.schedule(self._handler, path, recursive=True)

        def start(self, callback):
            """
            Start watching.
            """
            if not self._started:

                def on_any_event(evt):
                    event = {
                        'type': evt.event_type,
                        'abs_path': os.path.abspath(evt.src_path),
                        'rel_path': os.path.relpath(evt.src_path, self._working_dir),
                        'is_directory': evt.is_directory,
                    }

                    from twisted.internet import reactor
                    reactor.callFromThread(callback, event)

                self._handler.on_any_event = on_any_event
                self._observer.start()

        def stop(self):
            """
            Stop watching.
            """
            if self._started:
                self._observer.stop()
                self._observer.join()
                self._started = False

        def is_started(self):
            """
            Check if the watcher is running.
            """
            return self._started



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

__all__ = ('BUILTIN_ROLES', )

BUILTIN_ROLES = {
    u'node-role': {
        u'name':
        u'node',
        u'permissions': [{
            u'uri': u'',
            u'match': u'prefix',
            u'allow': {
                u'call': True,
                u'register': True,
                u'publish': True,
                u'subscribe': True
            },
            u'disclose': {
                u'caller': True,
                u'publisher': True
            },
            u'cache': True
        }]
    },
    u'backend-role': {
        u'name':
        u'backend',
        u'permissions': [{
            u'uri': u'',
            u'match': u'prefix',
            u'allow': {
                u'call': True,
                u'register': True,
                u'publish': True,
                u'subscribe': True
            },
            u'disclose': {
                u'caller': True,
                u'publisher': True
            },
            u'cache': True
        }]
    },
    u'public-role': {
        u'name':
        u'public',
        u'permissions': [{
            u'uri': u'crossbarfabriccenter.mrealm.get_status',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.get_nodes',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.remote.node.get_workers',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.remote.node.get_worker',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.remote.router.get_router_realms',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.remote.realm.meta.wamp.session.list',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.remote.realm.meta.wamp.session.get',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.remote.realm.meta.wamp.session.on_join',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.remote.realm.meta.wamp.session.on_leave',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.get_traces',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.get_trace',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.get_trace_data',
            u'allow': {
                u'call': True
            },
            u'disclose': {
                u'caller': True,
            },
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_data',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_created',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_starting',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_started',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_stopping',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_stopped',
            u'allow': {
                u'subscribe': True
            }
        }, {
            u'uri': u'crossbarfabriccenter.mrealm.tracing.on_trace_deleted',
            u'allow': {
                u'subscribe': True
            }
        }]
    },
    u'owner-role': {
        u'name':
        u'owner',
        u'permissions': [{
            u'uri': u'',
            u'match': u'prefix',
            u'allow': {
                u'call': True,
                u'register': True,
                u'publish': True,
                u'subscribe': True
            },
            u'disclose': {
                u'caller': True,
                u'publisher': True
            },
            u'cache': True
        }]
    }
}



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from crossbar.master.api.remote import RemoteApi

__all__ = ('RemoteNodeApi', )


class RemoteNodeApi(RemoteApi):
    """
    Remote API to CF node controller.
    """

    PREFIX = u'crossbarfabriccenter.remote.node.'

    PROCS = {
        # these are node level procedures
        u'node': [
            u'get_cpu_count',
            u'get_cpu_affinity',
            u'set_cpu_affinity',
            u'get_process_info',
            u'get_process_stats',
            u'set_process_stats_monitoring',
            u'trigger_gc',
            u'start_manhole',
            u'stop_manhole',
            u'get_manhole',
            u'get_status',
            u'get_system_stats',
            u'shutdown',
            u'get_worker_log',
            u'get_workers',
            u'get_worker',
            u'start_worker',
            u'stop_worker',
        ],
    }  # yapf: disable

    EVENTS = {
        # these are node level topics
        u'node': [
            u'on_node_starting',
            u'on_node_started',
            u'on_node_heartbeat',
            u'on_node_stopping',
            u'on_node_stopped',
            u'on_router_starting',
            u'on_router_started',
            u'on_container_starting',
            u'on_container_started',
            u'on_guest_starting',
            u'on_guest_started',
            u'on_proxy_starting',
            u'on_proxy_started',
            u'on_xbrmm_starting',
            u'on_xbrmm_started',
        ],
    }



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import txaio

txaio.use_twisted()  # noqa

import unittest
import crossbar.edge.node.docker as docker
from crossbar.edge.node.tests.dockerinspect import my_json
from unittest import mock

dummy_scandir = mock.MagicMock()
files_result = {'dirs': ['folder1', 'folder2'], 'files': ['dick', 'harry', 'tom']}
read_data = 'this is my data'
write_data = 'this is write data'


class DirEntry:
    def __init__(self, name, isFile):
        self._name = name
        self._isFile = isFile

    @property
    def name(self):
        return self._name

    def is_dir(self):
        return not self._isFile

    def is_file(self):
        return self._isFile


class TestDocker(unittest.TestCase):
    def setUp(self):

        # with open('tests/docker-inspect.json') as io:
        #     my_json = json.loads(io.read())

        class Attrs():
            attrs = my_json

        class MyEnv:
            containers = {0: Attrs()}

        def from_env():
            return MyEnv()

        self.client = docker.DockerClient(None, None)
        self.client._docker = mock.Mock()
        self.client._docker.from_env = from_env

        global dummy_scandir
        dummy_scandir().__enter__().__iter__.return_value = iter([
            DirEntry('tom', True),
            DirEntry('dick', True),
            DirEntry('harry', True),
            DirEntry('folder1', False),
            DirEntry('folder2', False),
        ])

    def test_fs_root(self):
        #
        #   We may only open files that are contained within volumes and as such are
        #   listed in 'Mounts'. By default these will be specifc RW volumes, OR, stuff
        #   we've configured to be editable using the env variable supplied in the UI.
        #
        assert self.client.fs_root(0, '/app/hello') == '/var/lib/docker/root/hello'
        assert self.client.fs_root(0, '/app/there') == '/var/lib/docker/root/there'
        #
        #   Access outside of a bind or mount point will throw an error
        #
        with self.assertRaises(Exception):
            self.client.fs_root(0, '/appx/crap')
        assert self.client.fs_root(0, '/home/hello') == '/var/lib/docker/home/hello'
        assert self.client.fs_root(0, '/home/there') == '/var/lib/docker/home/there'
        with self.assertRaises(Exception):
            self.client.fs_root(0, '/homex/crap')

    @mock.patch('os.scandir', dummy_scandir)
    def test_fs_open_app(self):
        #
        #   Check mount # 1
        #
        result = self.client.fs_open(0, '/app/myfiles')
        assert result == files_result

    @mock.patch('os.scandir', dummy_scandir)
    def test_fs_open_root(self):
        result = self.client.fs_open(0, '/')
        assert result == {'files': [], 'dirs': ['/app', '/home']}
        result = self.client.fs_open(0, '')
        assert result == {'files': [], 'dirs': ['/app', '/home']}

    @mock.patch('os.scandir', dummy_scandir)
    def test_fs_open_home(self):
        #
        #   Check mount # 2
        #
        result = self.client.fs_open(0, '/home/myfiles')
        assert result == files_result

    def test_fs_get_fail(self):
        raise unittest.SkipTest('FIXME: Fails on matterhorn, probably permissions issue.')
        with self.assertRaises(FileNotFoundError):
            self.client.fs_open(0, '/home/myfile')

    @mock.patch("builtins.open", mock.mock_open(read_data=read_data))
    def test_fs_get_ok(self):
        result = self.client.fs_get(0, '/home/myfile')
        assert result['data'] == read_data

    def test_fs_put_ok(self):
        with mock.patch("builtins.open", mock.mock_open()) as m:
            self.client.fs_put(0, '/home/myfile', write_data)
            m.assert_called_with('/var/lib/docker/home/myfile', 'w')
            m().write.assert_called_with(write_data)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from txaio import make_logger, time_ns

from crossbar.common.checkconfig import check_dict_args

__all__ = ('Monitor', )


class Monitor(object):
    """
    Host monitor base class.
    """

    ID = 'abstract'
    """
    Sensor ID, must defined in derived class.
    """

    log = make_logger()

    def __init__(self, config=None):
        """

        :param config: Submonitor specific configuration.
        :type config: dict or None
        """
        # submonitor specific configuration
        self._config = config

        # incremented on each poll
        self._tick = 0

        # time of last poll: ns Unix time UTC
        self._last_poll = None

        # effective period corresponding to last poll in ns
        self._last_period = None

        # last polled value
        self._last_value = None

        # total elapsed CPU time in ns reading this sensor
        self._elapsed = 0

    def check(self, config):
        """
        Check submonitor configuration item.

        Override in your derived submonitor class.

        Raise a `crossbar.common.checkconfig.InvalidConfigException` exception
        when you find an error in the item configuration.

        :param config: The submonitor configuration item to check.
        :type config: dict
        """
        check_dict_args({}, config, '{} monitor configuration'.format(self.ID))

    def poll(self):
        """
        Measure current stats value and return new stats.

        Override in your derived submonitor class.

        :returns: Current stats from monitor.
        :rtype: dict
        """
        self._tick += 1

        now = time_ns()
        if self._last_poll:
            self._last_period = now - self._last_poll

        current = {
            u'tick': self._tick,

            # the UTC timestamp when measurement was taken
            u'timestamp': now,

            # the effective last period in ns
            u'last_period': self._last_period,

            # duration in seconds the retrieval of sensor values took
            u'elapsed': self._elapsed,
        }

        self._last_poll = now
        self._last_value = current

        return current

    def get(self, details=None):
        """
        Get last stats/mesasurement values.

        Usually, there is no need to override this in a derived submonitor, as
        the default implementation already handles storing and returning the
        last submonitor reading.

        :returns: Last stats/values from monitor.
        :rtype: dict or None (when not yet polled)
        """
        self.log.info('{klass}.get(details={})', klass=self.__class__.__name__, details=details)
        return self._last_value



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from autobahn.wamp.exception import ApplicationError

from crossbar.test import TestCase
from crossbar.worker import _appsession_loader


class AppSessionLoaderTests(TestCase):
    """
    Tests for C{_appsession_loader}.
    """
    def test_unknown_types(self):
        """
        An unknown type will raise an exception.
        """
        config = {
            "type": "kjdfgdbfls",
        }

        with self.assertRaises(ApplicationError) as e:
            _appsession_loader(config)

        self.assertIn(("invalid component type 'kjdfgdbfls'"), str(e.exception.args[0]))

    def test_class_standard(self):
        """
        You can load a standard class.
        """
        config = {"type": "class", "classname": "crossbar.worker.test.examples.goodclass.AppSession"}

        klass = _appsession_loader(config)

        from .examples.goodclass import AppSession
        self.assertIs(klass, AppSession)

    def test_class_non_applicationsession(self):
        """
        Loading a class which does not subclass AppSession fails.
        """
        config = {"type": "class", "classname": "crossbar.worker.test.examples.badclass.AppSession"}

        with self.assertRaises(ApplicationError) as e:
            _appsession_loader(config)

        self.assertIn(("session not derived of ApplicationSession"), str(e.exception.args[0]))

    def test_class_importerror(self):
        """
        Loading a class which has an import error upon import raises that
        error.
        """
        config = {"type": "class", "classname": "crossbar.worker.test.examples.importerror.AppSession"}

        with self.assertRaises(ApplicationError) as e:
            _appsession_loader(config)

        self.assertIn(("Failed to import class 'crossbar.worker.test.examples.importerr"
                       "or.AppSession'"), str(e.exception.args[0]))

        s = str(e.exception.args[0])
        self.assertTrue('ImportError' in s or 'ModuleNotFoundError' in s)

    def test_class_syntaxerror(self):
        """
        Loading a class which has a SyntaxError raises that up.
        """
        config = {"type": "class", "classname": "crossbar.worker.test.examples.syntaxerror.AppSession"}

        with self.assertRaises(ApplicationError) as e:
            _appsession_loader(config)

        self.assertIn(("Failed to import class 'crossbar.worker.test.examples.syntaxerr"
                       "or.AppSession'"), str(e.exception.args[0]))
        self.assertIn(("SyntaxError"), str(e.exception.args[0]))



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
import psutil

from crossbar.edge.worker.monitor._base import Monitor
from txaio import perf_counter_ns

__all__ = ('SystemMonitor', )


class SystemMonitor(Monitor):
    """
    System monitoring via psutils.
    """

    ID = u'system'

    def poll(self):
        hdata = Monitor.poll(self)

        start = perf_counter_ns()

        hdata['cpu'] = psutil.cpu_times_percent()._asdict()
        hdata['memory'] = psutil.virtual_memory()._asdict()
        hdata['loadavg'] = os.getloadavg()

        # uptime, as all durations, is in ns
        hdata['uptime'] = int(hdata['timestamp'] - psutil.boot_time() * 10**10)

        hdata[u'elapsed'] = perf_counter_ns() - start

        self._last_value = hdata

        return hdata
